import json
import logging
import socket
from contextlib import contextmanager, asynccontextmanager
from copy import deepcopy
from datetime import datetime
from pathlib import Path
from typing import MutableMapping, Any, Literal, Generator, Coroutine, Callable, Awaitable, Unpack

from aiohttp import ClientSession, ClientResponse
from yarl import URL

from aiorequestful.exception import AuthoriserError
from aiorequestful.types import MethodInput, URLInput, Method, Headers, ImmutableHeaders, MutableJSON, ImmutableJSON, \
    JSON, RequestKwargs


class AuthRequest:
    """
    Request handler for sending authentication and authorisation requests.
    Supply this class with the required arguments for your request.

    :param method: HTTP request method (such as GET, POST, PUT, etc.).
    :param url: The URL of the request.
    :param **kwargs: Any other kwargs required for a successful request.
    """

    def __init__(self, method: MethodInput, url: URLInput, **kwargs: Unpack[RequestKwargs]):
        self.method = Method.get(method)
        self.url = URL(url)

        for k, v in kwargs.items():
            setattr(self, k, v)

    @classmethod
    def _sanitise_kwargs(cls, kwargs: MutableMapping[str, Any]) -> None:
        cls._sanitise_map(kwargs.get("params"))
        cls._sanitise_map(kwargs.get("data"))

    @classmethod
    def _sanitise_map(cls, value: MutableMapping[str, Any] | None) -> None:
        if not value:
            return

        for k, v in value.items():
            if isinstance(v, MutableMapping):
                cls._sanitise_map(v)
            elif isinstance(v, bool) or not isinstance(v, str | int | float):
                value[k] = json.dumps(v)

    @contextmanager
    def enrich_parameters(
            self, key: Literal["data", "params", "json"], value: dict[str, Any]
    ) -> Generator[None, None, None]:
        """
        Temporarily append data to the parameters of a request and remove when done.

        :param key: The keyword of the argument to append data to.
        :param value: The value to append.
        """
        current_value = getattr(self, key, {})
        setattr(self, key, current_value | value)

        yield

        if current_value:
            setattr(self, key, current_value)
        else:
            delattr(self, key)

    def __call__(self, session: ClientSession) -> Coroutine[ClientResponse, None, None]:
        return self.request(session=session)

    @asynccontextmanager
    async def request(self, session: ClientSession) -> Coroutine[ClientResponse, None, None]:
        """Send the request within the given ``session`` and return the response."""
        kwargs = {k: deepcopy(v) for k, v in vars(self).items() if k not in ("method", "url")}
        self._sanitise_kwargs(kwargs)

        async with session.request(method=self.method.name, url=self.url, **kwargs) as response:
            yield response


class AuthResponseHandler:
    """
    Handle saving, loading, enriching, sanitising etc. of responses.
    Also handles token extraction and header generation from token responses.

    Ideally, usage of this class should ensure that the stored response is valid i.e.
    a response should not be stored on this class if it is invalid.

    :param file_path: Path to use for loading and saving a token.
    :param token_prefix_default: Prefix to add to the header value for authorised calls to an endpoint.
    :param additional_headers: Extra headers to add to the final headers to ensure future successful requests.
    """

    @property
    def token(self) -> str:
        """Extract the token from the stored response."""
        if not self.response:
            raise AuthoriserError("Stored response is not available.")

        if self.token_key not in self.response:
            raise AuthoriserError(
                f"Did not find valid token at key: {self.token_key} | {self.sanitise_response()}"
            )
        return str(self.response[self.token_key])

    @property
    def headers(self) -> Headers:
        """Generate headers from the stored response, adding all additional headers as needed."""
        header_key = "Authorization"
        header_prefix = self.response.get("token_type", self.token_prefix_default)

        headers = {header_key: f"{header_prefix} {self.token}"}
        if self.additional_headers:
            headers.update(self.additional_headers)
        return headers

    def __init__(
            self,
            file_path: str | Path = None,
            token_prefix_default: str | None = None,
            additional_headers: ImmutableHeaders = None,
    ):
        #: The :py:class:`logging.Logger` for this  object
        self.logger: logging.Logger = logging.getLogger(__name__)

        self.response: MutableJSON | None = None

        self.file_path: Path | None = Path(file_path).with_suffix(".json") if file_path else None
        self.token_key: str = "access_token"
        self.token_prefix_default: str | None = token_prefix_default

        self.additional_headers = additional_headers

    def sanitise_response(self, response: ImmutableJSON = None) -> JSON:
        """
        Returns a reformatted ``response``, making it safe to log by removing sensitive values at predefined keys.
        If no ``response`` is given, uses the stored response.
        """
        response = response or self.response
        if not response:
            return {}

        def _clean_value(value: Any) -> str:
            value = str(value)
            if len(value) < 5:
                return ""
            return f"{value[:5]}..."

        response_clean = {k: _clean_value(v) if str(k).endswith("_token") else v for k, v in response.items()}
        if self.token_key in response_clean:
            response_clean[self.token_key] = _clean_value(response_clean[self.token_key])

        return response_clean

    def enrich_response(self, response: MutableJSON = None, refresh_token: str = None) -> None:
        """
        Extends the ``response`` by adding granted and expiry time information to it.
        Adds the given ``refresh_token`` to the response if one is not present.
        If no ``response`` is given, uses the stored response.
        """
        response = response or self.response
        if not response:
            return

        # add granted and expiry times to token
        response["granted_at"] = datetime.now().timestamp()
        if "expires_in" in response:
            expires_at = response["granted_at"] + float(response["expires_in"])
            response["expires_at"] = expires_at

        # request usually does return a new refresh token, but add the previous one if not
        if "refresh_token" not in response and refresh_token:
            response["refresh_token"] = refresh_token

    def load_response_from_file(self) -> JSON | None:
        """Load a stored response from given path"""
        if not self.file_path or not self.file_path.exists():
            return

        self.logger.debug("Saved authorisation code response found. Loading...")
        with open(self.file_path, "r") as file:  # load token
            self.response = json.load(file)

        return self.response

    def save_response_to_file(self, response: MutableJSON = None) -> None:
        """
        Save a ``response`` to given path.
        If a ``response`` is given, updates the stored response.
        If no ``response`` is given, saves the stored response.
        """
        self.response = response or self.response
        if not self.file_path or not self.response:
            return

        self.logger.debug(f"Saving authorisation code response: {self.sanitise_response()}")
        with open(self.file_path, "w") as file:
            json.dump(self.response, file, indent=2)


class AuthResponseTester:
    """
    Run tests against the response of authorisation request to ensure its validity.

    When setting ``max_expiry``, the following example illustrates how this is used:
        * A token has 600 second total expiry time,
        * it is 60 seconds old and therefore still has 540 seconds of authorised time left,
        * you set ``max_expiry`` = 300, the token will pass tests.
        * The same token is tested again later when it is 500 now seconds old,
        * it now has only 100 seconds of authorised time left,
        * it will now fail the tests as 100 < 300.

    :param request: The request to execute when testing the access token.
    :param response_test: Test to apply to the response from the access token request.
    :param max_expiry: The max allowed time in seconds left until the token is due to expire.
        Useful for ensuring the token will be valid for long enough to run your operations.
    """
    def __init__(
            self,
            request: AuthRequest | None = None,
            response_test: Callable[[ClientResponse], Awaitable[bool]] | None = None,
            max_expiry: int = 0,
    ):
        #: The :py:class:`logging.Logger` for this  object
        self.logger: logging.Logger = logging.getLogger(__name__)

        self.request = request
        self.response_test = response_test
        #: The max allowed time in seconds left until the token is due to expire
        self.max_expiry = max_expiry

    def __call__(
            self, response: ImmutableJSON | None = None, headers: ImmutableHeaders | None = None
    ) -> Awaitable[bool]:
        return self.test(response=response, headers=headers)

    async def test(self, response: ImmutableJSON | None = None, headers: ImmutableHeaders | None = None) -> bool:
        """Test validity of the ``response`` and given ``headers``. Returns True if all tests pass, False otherwise"""
        if not response:
            return False

        self.logger.debug("Begin testing auth response...")

        result = self._test_response(response=response)
        if result:
            result = self._test_expiry(response=response)
        if result:
            result = await self._test_token(headers=headers)

        return result

    def _test_response(self, response: ImmutableJSON) -> bool:
        result = "error" not in response
        self.logger.debug(f"Auth response contains no error test: {result}")
        return result

    def _test_expiry(self, response: ImmutableJSON) -> bool:
        if all(key not in response for key in ("expires_at", "expires_in")) or self.max_expiry <= 0:
            return True

        if "expires_at" in response:
            result = datetime.now().timestamp() + self.max_expiry < response["expires_at"]
        else:
            result = self.max_expiry < response["expires_in"]

        self.logger.debug(f"Token expiry time test: {result}")
        return result

    async def _test_token(self, headers: ImmutableHeaders | None) -> bool:
        if self.request is None or self.response_test is None:
            return True

        with self.request.enrich_parameters("headers", headers):
            async with ClientSession() as session:
                async with self.request(session=session) as response:
                    result = await self.response_test(response)

        self.logger.debug(f"Validate token test: {result}")
        return result if result is not None else False


class SocketHandler:
    """
    :param port: The port to open on the localhost for this socket.
    :param timeout: The time in seconds to keep the socket listening for a request.
    """

    def __init__(self, port: int = 8080, timeout: int = 120):
        #: The port to open on the localhost for this socket
        self.port = port
        #: The time in seconds to keep the socket listening for a request.
        self.timeout = timeout

        self._socket: socket.socket | None = None

    def __enter__(self) -> socket.socket:
        """Set up socket to listen for the callback"""
        self._socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        self._socket.bind(("localhost", self.port))
        self._socket.settimeout(self.timeout)
        self._socket.listen(1)
        return self._socket

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._socket.close()



from __future__ import annotations

import json
import os
from collections.abc import Mapping, Callable, Generator
from datetime import datetime, timedelta
from pathlib import Path
from tempfile import gettempdir
from typing import Any, Self

from aiohttp import RequestInfo, ClientRequest, ClientResponse
from dateutil.relativedelta import relativedelta

from aiorequestful import PROGRAM_NAME
from aiorequestful._utils import required_modules_installed
from aiorequestful.cache.backend.base import DEFAULT_EXPIRE, ResponseCache, ResponseRepository, RepositoryRequestType
from aiorequestful.cache.backend.base import RequestSettings
from aiorequestful.exception import CacheError
from aiorequestful.types import URLInput

try:
    import aiosqlite
except ImportError:
    aiosqlite = None


class SQLiteTable[K: tuple[Any, ...], V: str](ResponseRepository[K, V]):

    __slots__ = ()

    #: The column under which a response's name is stored in the table
    name_column = "name"
    #: The column under which response data is stored in the table
    data_column = "response"
    #: The column under which the response cache time is stored in the table
    cached_column = "cached_at"
    #: The column under which the response expiry time is stored in the table
    expiry_column = "expires_at"

    # noinspection PyPropertyDefinition
    @classmethod
    @property
    def _required_modules(cls) -> list:
        return [aiosqlite]

    async def create(self) -> Self:
        ddl_sep = "\t, "
        ddl = "\n".join((
            f'CREATE TABLE IF NOT EXISTS "{self.settings.name}" (',
            '\t' + f'\n{ddl_sep}'.join(
                f'"{key}" {data_type} NOT NULL' for key, data_type in self._primary_key_columns.items()
            ),
            f'{ddl_sep}"{self.name_column}" TEXT',
            f'{ddl_sep}"{self.cached_column}" TIMESTAMP NOT NULL',
            f'{ddl_sep}"{self.expiry_column}" TIMESTAMP NOT NULL',
            f'{ddl_sep}"{self.data_column}" TEXT',
            f'{ddl_sep}PRIMARY KEY ("{'", "'.join(self._primary_key_columns)}")',
            ');',
            f'CREATE INDEX IF NOT EXISTS idx_{self.expiry_column} '
            f'ON "{self.settings.name}"({self.expiry_column});'
        ))

        self.logger.debug(f"Creating {self.settings.name!r} table with the following DDL:\n{ddl}")
        await self.connection.executescript(ddl)
        await self.commit()

        return self

    def __init__(
            self,
            connection: aiosqlite.Connection,
            settings: RequestSettings,
            expire: timedelta | relativedelta = DEFAULT_EXPIRE,
    ):
        required_modules_installed(self._required_modules, self)

        super().__init__(settings=settings, expire=expire)

        self.connection = connection

    def __await__(self) -> Generator[Any, None, Self]:
        return self.create().__await__()

    async def commit(self) -> None:
        """Commit the transactions to the database."""
        try:
            await self.connection.commit()
        except ValueError:
            pass

    async def close(self) -> None:
        try:
            await self.commit()
            await self.connection.close()
        except ValueError:
            pass

    @property
    def _primary_key_columns(self) -> Mapping[str, str]:
        """A map of column names to column data types for the primary keys of this repository."""
        expected_columns = self.settings.fields

        keys = {"method": "VARCHAR(10)"}
        if "id" in expected_columns:
            keys["id"] = "VARCHAR(50)"
        if "offset" in expected_columns:
            keys["offset"] = "INT2"
        if "size" in expected_columns:
            keys["size"] = "INT2"

        return keys

    def get_key_from_request(self, request: RepositoryRequestType[K]) -> K | None:
        if isinstance(request, ClientRequest | ClientResponse):
            request = request.request_info
        if not isinstance(request, RequestInfo):
            return request  # `request` is the key

        key = self.settings.get_key(request.url)
        if any(part is None for part in key):
            return

        return str(request.method).upper(), *key

    async def count(self, include_expired: bool = True) -> int:
        query = f'SELECT COUNT(*) FROM "{self.settings.name}"'
        params = []

        if not include_expired:
            query += f'\nWHERE "{self.expiry_column}" > ?'
            params.append(datetime.now().isoformat())

        async with self.connection.execute(query, params) as cur:
            row = await cur.fetchone()

        return row[0]

    async def contains(self, request: RepositoryRequestType[K]) -> bool:
        key = self.get_key_from_request(request)
        query = "\n".join((
            f'SELECT COUNT(*) FROM "{self.settings.name}"',
            f'WHERE "{self.expiry_column}" > ?',
            f'\tAND {'\n\tAND '.join(f'"{key}" = ?' for key in self._primary_key_columns)}',
        ))

        async with self.connection.execute(query, (datetime.now().isoformat(), *key)) as cur:
            rows = await cur.fetchone()
        return rows[0] > 0

    async def clear(self, expired_only: bool = False) -> int:
        query = f'DELETE FROM "{self.settings.name}"'
        params = []

        if expired_only:
            query += f'\nWHERE "{self.expiry_column}" > ?'
            params.append(datetime.now().isoformat())

        async with self.connection.execute(query, params) as cur:
            count = cur.rowcount
        return count

    async def __aiter__(self):
        query = "\n".join((
            f'SELECT "{'", "'.join(self._primary_key_columns)}", "{self.data_column}" ',
            f'FROM "{self.settings.name}"',
            f'WHERE "{self.expiry_column}" > ?',
        ))
        async with self.connection.execute(query, (datetime.now().isoformat(),)) as cur:
            async for row in cur:
                yield row[:-1], self.deserialize(row[-1])

    async def get_response(self, request: RepositoryRequestType[K]) -> V | None:
        key = self.get_key_from_request(request)
        if not key:
            return

        query = "\n".join((
            f'SELECT "{self.data_column}" FROM {self.settings.name}',
            f'WHERE "{self.data_column}" IS NOT NULL',
            f'\tAND "{self.expiry_column}" > ?',
            f'\tAND {'\n\tAND '.join(f'"{key}" = ?' for key in self._primary_key_columns)}',
        ))

        async with self.connection.execute(query, (datetime.now().isoformat(), *key)) as cur:
            row = await cur.fetchone()

        if not row:
            return
        return self.deserialize(row[0])

    async def _set_item_from_key_value_pair(self, __key: K, __value: Any) -> None:
        columns = (
            *self._primary_key_columns,
            self.name_column,
            self.cached_column,
            self.expiry_column,
            self.data_column
        )
        query = "\n".join((
            f'INSERT OR REPLACE INTO "{self.settings.name}" (',
            f'\t"{'", "'.join(columns)}"',
            ') ',
            f'VALUES({','.join('?' * len(columns))});',
        ))
        params = (
            *__key,
            self.settings.get_name(self.deserialize(__value)),
            datetime.now().isoformat(),
            self.expire.isoformat(),
            self.serialize(__value)
        )

        await self.connection.execute(query, params)

    async def delete_response(self, request: RepositoryRequestType[K]) -> bool:
        key = self.get_key_from_request(request)
        query = "\n".join((
            f'DELETE FROM "{self.settings.name}"',
            f'WHERE {'\n\tAND '.join(f'"{key}" = ?' for key in self._primary_key_columns)}',
        ))

        async with self.connection.execute(query, key) as cur:
            count = cur.rowcount
        return count > 0

    def serialize(self, value: Any) -> V | None:
        if isinstance(value, str):
            try:
                value = json.loads(value)
            except json.decoder.JSONDecodeError:
                return

        return json.dumps(value, indent=2)

    def deserialize(self, value: V | dict) -> Any:
        if isinstance(value, dict):
            return value

        try:
            return json.loads(value)
        except (json.decoder.JSONDecodeError, TypeError):
            return


class SQLiteCache(ResponseCache[SQLiteTable]):

    __slots__ = ("_connector", "connection")

    # noinspection PyPropertyDefinition
    @classmethod
    @property
    def type(cls):
        return "sqlite"

    @property
    def closed(self):
        """Is the stored client session closed."""
        return self.connection is None or not self.connection.is_alive()

    @staticmethod
    def _get_sqlite_path(path: Path) -> Path:
        return path.with_suffix(".sqlite")

    @staticmethod
    def _clean_kwargs[T: dict](kwargs: T) -> T:
        kwargs.pop("cache_name", None)
        kwargs.pop("connection", None)
        return kwargs

    @classmethod
    def connect(cls, value: Any, **kwargs) -> Self:
        return cls.connect_with_path(path=value, **kwargs)

    @classmethod
    def connect_with_path(cls, path: os.PathLike, **kwargs) -> Self:
        """Connect with an SQLite DB at the given ``path`` and return an instantiated :py:class:`SQLiteResponseCache`"""
        path = cls._get_sqlite_path(Path(path))
        os.makedirs(path.parent, exist_ok=True)

        return cls(
            cache_name=str(path),
            connector=lambda: aiosqlite.connect(database=path),
            **cls._clean_kwargs(kwargs)
        )

    @classmethod
    def connect_with_in_memory_db(cls, **kwargs) -> Self:
        """Connect with an in-memory SQLite DB and return an instantiated :py:class:`SQLiteResponseCache`"""
        return cls(
            cache_name="__IN_MEMORY__",
            connector=lambda: aiosqlite.connect(database="file::memory:?cache=shared", uri=True),
            **cls._clean_kwargs(kwargs)
        )

    @classmethod
    def connect_with_temp_db(cls, name: str = f"{PROGRAM_NAME.lower()}_db.tmp", **kwargs) -> Self:
        """Connect with a temporary SQLite DB and return an instantiated :py:class:`SQLiteResponseCache`"""
        path = cls._get_sqlite_path(Path(gettempdir(), name))
        return cls(
            cache_name=name,
            connector=lambda: aiosqlite.connect(database=path),
            **cls._clean_kwargs(kwargs)
        )

    def __init__(
            self,
            cache_name: str,
            connector: Callable[[], aiosqlite.Connection],
            repository_getter: Callable[[Self, URLInput], SQLiteTable] = None,
            expire: timedelta | relativedelta = DEFAULT_EXPIRE,
    ):
        # noinspection PyProtectedMember,PyTypeChecker
        required_modules_installed(SQLiteTable._required_modules, self)

        super().__init__(cache_name=cache_name, repository_getter=repository_getter, expire=expire)

        self._connector = connector
        self.connection: aiosqlite.Connection | None = None

    async def _connect(self) -> Self:
        if self.closed:
            self.connection = self._connector()
            await self.connection

        for repository in self._repositories.values():
            repository.connection = self.connection
            await repository.create()

        return self

    def __await__(self) -> Generator[Any, None, Self]:
        return self._connect().__await__()

    async def __aexit__(self, __exc_type, __exc_value, __traceback) -> None:
        if self.closed:
            return

        await self.commit()
        await self.connection.__aexit__(__exc_type, __exc_value, __traceback)
        self.connection = None

    async def commit(self):
        """Commit the transactions to the database."""
        if self.closed:
            return

        try:
            await self.connection.commit()
        except ValueError:
            pass

    async def close(self):
        if self.closed:
            return

        try:
            await self.commit()
            await self.connection.close()
        except ValueError:
            pass

    def create_repository(self, settings: RequestSettings) -> SQLiteTable:
        if settings.name in self:
            raise CacheError(f"Repository already exists: {settings.name}")

        repository = SQLiteTable(connection=self.connection, settings=settings, expire=self.expire)
        self._repositories[settings.name] = repository
        return repository



"""
Exceptions relating to API operations.
"""
from typing import Any

from aiohttp import ClientResponse


class AIORequestsError(Exception):
    """Generic base class for all aiorequestful-related errors"""


class InputError(AIORequestsError, ValueError):
    """Exception raised when the given input is invalid."""


class MethodError(InputError):
    """Exception raised for unrecognised method values."""
    def __init__(self, value: Any = None, message: str = "Unrecognised method"):
        self.value = value
        if value is not None:
            message = f"{message}: {value}"
        super().__init__(message)


class AIORequestsImportError(AIORequestsError, ImportError):
    """Exception raised for import errors, usually from missing modules."""


###########################################################################
## HTTP Requests errors
###########################################################################
class HTTPError(AIORequestsError):
    """Exception raised for generic HTTP errors."""


class RequestError(HTTPError):
    """Exception raised for errors relating to HTTP requests."""


class ResponseError(HTTPError):
    """
    Exception raised for errors relating to responses from HTTP requests.

    :param message: Explanation of the error.
    :param response: The :py:class:`ClientResponse` related to the error.
    """
    def __init__(self, message: str = None, response: ClientResponse | None = None):
        self.message = message
        self.response = response
        formatted = f"Status code: {response.status} | {message}" if response else message
        super().__init__(formatted)


###########################################################################
## Authoriser errors
###########################################################################
class AuthoriserError(AIORequestsError):
    """Exception raised for errors relating to authorisation."""


###########################################################################
## Cache errors
###########################################################################
class CacheError(AIORequestsError):
    """Exception raised for errors relating to the cache."""



import logging
from abc import ABC, abstractmethod
from collections.abc import Awaitable

from aiorequestful.types import Headers


_DEFAULT_SERVICE_NAME = "unknown_service"


class Authoriser(ABC):
    """
    Base interface for authenticating and authorising access to a service over HTTP.

    :param service_name: The service name for which to authorise.
    """

    def __init__(self, service_name: str = _DEFAULT_SERVICE_NAME):
        #: The :py:class:`logging.Logger` for this  object
        self.logger: logging.Logger = logging.getLogger(__name__)

        #: The service name for which to authorise. Currently only used for logging purposes.
        self.service_name = service_name

    @abstractmethod
    async def authorise(self) -> Headers:
        """
        Authenticate and authorise, testing/refreshing/re-authorising as needed.

        :raise APIError: If the token cannot be generated or validated.
        """
        raise NotImplementedError

    def __call__(self) -> Awaitable[Headers]:
        return self.authorise()



"""
All operations relating to handling of requests to an API.
"""
import asyncio
import contextlib
import inspect
import json
import logging
from collections.abc import Mapping, Callable
from datetime import datetime, timedelta
from http import HTTPStatus
from typing import Any, Self, Unpack
from urllib.parse import unquote

import aiohttp
from aiohttp import ClientResponse, ClientSession
from yarl import URL

from aiorequestful._utils import format_url_log
from aiorequestful.auth import Authoriser
from aiorequestful.cache.backend import ResponseCache
from aiorequestful.cache.session import CachedSession
from aiorequestful.exception import RequestError, ResponseError

from aiorequestful.types import JSON, URLInput, Headers, MethodInput, Method, RequestKwargs


class RequestHandler:
    """
    Generic API request handler using cached responses for GET requests only.
    Caches GET responses for a maximum of 4 weeks by default.
    Handles error responses and backoff on failed requests.
    See :py:class:`APIAuthoriser` for more info on which params to pass to authorise requests.

    :param connector: When called, returns a new session to use when making requests.
    :param authoriser: The authoriser to use when authorising requests to the API.
    """

    __slots__ = (
        "logger",
        "_connector",
        "_session",
        "authoriser",
        "backoff_start",
        "backoff_factor",
        "backoff_count",
        "_backoff_start_logged",
        "wait_time",
        "wait_increment",
        "wait_max",
        "_wait_start_logged",
    )

    @property
    def backoff_final(self) -> float:
        """
        The maximum wait time to retry a request in seconds
        until giving up when applying backoff to failed requests
        """
        return self.backoff_start * self.backoff_factor ** self.backoff_count

    @property
    def timeout(self) -> int:
        """The cumulative sum of all backoff intervals up to the final backoff time"""
        return int(sum(self.backoff_start * self.backoff_factor ** i for i in range(self.backoff_count + 1)))

    @property
    def closed(self):
        """Is the stored client session closed."""
        return self._session is None or self._session.closed

    @property
    def session(self) -> ClientSession:
        """The :py:class:`ClientSession` object if it exists and is open."""
        if not self.closed:
            return self._session

    @classmethod
    def create(cls, authoriser: Authoriser | None = None, cache: ResponseCache | None = None, **session_kwargs):
        """Create a new :py:class:`RequestHandler` with an appropriate session ``connector`` given the input kwargs"""
        def connector() -> ClientSession:
            """Create an appropriate session ``connector`` given the input kwargs"""
            if cache is not None:
                return CachedSession(cache=cache, **session_kwargs)
            return ClientSession(**session_kwargs)

        return cls(connector=connector, authoriser=authoriser)

    def __init__(self, connector: Callable[[], ClientSession], authoriser: Authoriser | None = None):
        #: The :py:class:`logging.Logger` for this  object
        self.logger: logging.Logger = logging.getLogger(__name__)

        self._connector = connector
        self._session: ClientSession | CachedSession | None = None

        #: The :py:class:`Authoriser` object
        self.authoriser = authoriser

        #: The initial backoff time in seconds for failed requests
        self.backoff_start = 0.2
        #: The factor by which to increase backoff time for failed requests i.e. backoff_start ** backoff_factor
        self.backoff_factor = 1.932
        #: The maximum number of request attempts to make before giving up and raising an exception
        self.backoff_count = 10
        self._backoff_start_logged = False

        #: The initial time in seconds to wait after receiving a response from a request
        self.wait_time = 0
        #: The amount in seconds to increase the wait time by each time a rate limit is hit i.e. 429 response
        self.wait_increment = 0.1
        #: The maximum time in seconds that the wait time can be incremented to
        self.wait_max = 1
        self._wait_start_logged = False

    async def __aenter__(self) -> Self:
        if self.closed:
            self._session = self._connector()

        await self.session.__aenter__()
        await self.authorise()

        return self

    async def __aexit__(self, __exc_type, __exc_value, __traceback) -> None:
        await self.session.__aexit__(__exc_type, __exc_value, __traceback)
        self._session = None

    async def authorise(self) -> Headers:
        """
        Method for API authorisation which tests/refreshes/reauthorises as needed.

        :return: Headers for request authorisation.
        :raise APIError: If the token cannot be validated.
        """
        if self.closed:
            raise RequestError("Session is closed. Enter the API context to start a new session.")

        headers = {}
        if self.authoriser is not None:
            self.session.headers.update(await self.authoriser())

        return headers

    async def close(self) -> None:
        """Close the current session. No more requests will be possible once this has been called."""
        await self.session.close()

    async def request(self, method: MethodInput, url: URLInput, **kwargs: Unpack[RequestKwargs]) -> JSON:
        """
        Generic method for handling API requests with back-off on failed requests.
        See :py:func:`request` for more arguments.

        :param method: method for the request:
            ``GET``, ``OPTIONS``, ``HEAD``, ``POST``, ``PUT``, ``PATCH``, or ``DELETE``.
        :param url: URL to call.
        :return: The JSON formatted response or, if JSON formatting not possible, the text response.
        :raise APIError: On any logic breaking error/response.
        """
        method = Method.get(method)
        backoff = self.backoff_start

        while True:
            async with self._request(method=method, url=url, **kwargs) as response:
                if response is None:
                    raise RequestError("No response received")

                handled = await self._handle_bad_response(response=response)
                waited = await self._wait_for_rate_limit_timeout(response=response)

                if handled or waited:
                    continue

                if response.ok:
                    data = await self._get_json_response(response)
                    break

                await self._log_response(response=response, method=method, url=url)

                if backoff > self.backoff_final or backoff == 0:
                    raise RequestError("Max retries exceeded")

                # exponential backoff
                self._log_backoff_start()
                self.log(method=method.name, url=url, message=f"Request failed: retrying in {backoff} seconds...")
                await asyncio.sleep(backoff)
                backoff *= self.backoff_factor

        self._backoff_start_logged = False
        return data

    @contextlib.asynccontextmanager
    async def _request(
            self,
            method: Method,
            url: URLInput,
            log_message: str | list[str] = None,
            **kwargs
    ) -> ClientResponse | None:
        """Handle logging a request, send the request, and return the response"""
        if isinstance(log_message, str):
            log_message = [log_message]
        elif log_message is None:
            log_message = []

        if isinstance(self.session, CachedSession):
            log_message.append("Cached Request")
        self.log(method=method.name, url=url, message=log_message, **kwargs)

        if not isinstance(self.session, CachedSession):
            self._clean_requests_kwargs(kwargs)
        if "headers" in kwargs:
            kwargs["headers"].update(self.session.headers)

        try:
            async with self.session.request(method=method.name, url=url, **kwargs) as response:
                yield response
                await asyncio.sleep(self.wait_time)
        except aiohttp.ClientError as ex:
            self.logger.debug(str(ex))
            yield

    @staticmethod
    def _clean_requests_kwargs(kwargs: dict[str, Any]) -> None:
        """Clean ``kwargs`` by removing any kwarg not in the signature of the :py:meth:`aiohttp.request` method."""
        signature = inspect.signature(aiohttp.request).parameters
        for key in list(kwargs):
            if key not in signature:
                kwargs.pop(key)

    def log(
            self, method: str, url: URLInput, message: str | list = None, level: int = logging.DEBUG, **kwargs
    ) -> None:
        """Format and log a request or request adjacent message to the given ``level``."""
        log: list[Any] = []

        url = URL(url)
        if url.query:
            log.extend(f"{k}: {unquote(v):<4}" for k, v in sorted(url.query.items()))
        if kwargs.get("params"):
            log.extend(f"{k}: {v:<4}" for k, v in sorted(kwargs.pop("params").items()))
        if kwargs.get("json"):
            log.extend(f"{k}: {str(v):<4}" for k, v in sorted(kwargs.pop("json").items()))
        if len(kwargs) > 0:
            log.extend(f"{k.title()}: {str(v):<4}" for k, v in kwargs.items() if v)
        if message:
            log.append(message) if isinstance(message, str) else log.extend(message)

        self.logger.log(level=level, msg=format_url_log(method=method, url=url, messages=log))

    def _log_backoff_start(self) -> None:
        if self._backoff_start_logged:
            return

        self.logger.warning(
            "\33[93mRequest failed: retrying using backoff strategy. "
            f"Will retry request {self.backoff_count} more times and timeout in {self.timeout} seconds...\33[0m"
        )
        self._backoff_start_logged = True

    async def _log_response(self, response: ClientResponse, method: Method, url: URLInput) -> None:
        """Log the method, URL, response text, and response headers."""
        response_headers = response.headers
        if isinstance(response.headers, Mapping):  # format headers if JSON
            response_headers = json.dumps(dict(response.headers), indent=2)
        self.log(
            method=method.name,
            url=url,
            message=[
                f"Status code: {response.status}",
                "Response text and headers follow:\n"
                f"Response text:\n\t{(await response.text()).replace("\n", "\n\t")}\n"
                f"Headers:\n\t{response_headers.replace("\n", "\n\t")}"
                f"\33[0m"
            ]
        )

    # TODO: Separate responsibility of handling specific status codes to implementations of a generic abstraction,
    #  and have this handling dependency injected for this handler.
    #  This doesn't follow SOLID very well;
    #  would need to modify function directly to implement handling of new status codes.
    async def _handle_bad_response(self, response: ClientResponse) -> bool:
        """Handle bad responses by extracting message and handling status codes that should raise an exception."""
        response_json = await self._get_json_response(response)
        error_message = response_json.get("error", {}).get("message")
        if error_message is None:
            status = HTTPStatus(response.status)
            error_message = f"{status.phrase} | {status.description}"

        handled = False

        def _log_bad_response(message: str) -> None:
            self.logger.debug(f"Status code: {response.status} | {error_message} | {message}")

        if response.status == 401:
            _log_bad_response("Re-authorising...")
            await self.authorise()
            handled = True
        elif response.status == 429:
            if self.wait_time < self.wait_max:
                self.wait_time += self.wait_increment
                _log_bad_response(f"Rate limit hit. Increasing wait time between requests to {self.wait_time}s")
                handled = True
            else:
                _log_bad_response(f"Rate limit hit and wait time already at maximum of {self.wait_time}s")
        elif 400 <= response.status < 408:
            raise ResponseError(error_message, response=response)

        return handled

    async def _wait_for_rate_limit_timeout(self, response: ClientResponse) -> bool:
        """Handle rate limits when a 'retry-after' time is included in the response headers."""
        if "retry-after" not in response.headers:
            return False

        wait_time = int(response.headers["retry-after"])
        wait_str = (datetime.now() + timedelta(seconds=wait_time)).strftime("%Y-%m-%d %H:%M:%S")

        if wait_time > self.timeout:  # exception if too long
            raise RequestError(
                f"Rate limit exceeded and wait time is greater than timeout of {self.timeout} seconds. "
                f"Retry again at {wait_str}"
            )

        if not self._wait_start_logged:
            self.logger.warning(f"\33[93mRate limit exceeded. Retrying again at {wait_str}\33[0m")
            self._wait_start_logged = True

        await asyncio.sleep(wait_time)
        self._wait_start_logged = False

        return True

    @staticmethod
    async def _get_json_response(response: ClientResponse) -> JSON:
        """Format the response to JSON and handle any errors"""
        try:
            data = await response.json()
            return data if isinstance(data, dict) else {}
        except (aiohttp.ContentTypeError, json.decoder.JSONDecodeError):
            return {}

    async def get(self, url: URLInput, **kwargs) -> JSON:
        """Sends a GET request."""
        kwargs.pop("method", None)
        return await self.request("get", url=url, **kwargs)

    async def post(self, url: URLInput, **kwargs) -> JSON:
        """Sends a POST request."""
        kwargs.pop("method", None)
        return await self.request("post", url=url, **kwargs)

    async def put(self, url: URLInput, **kwargs) -> JSON:
        """Sends a PUT request."""
        kwargs.pop("method", None)
        return await self.request("put", url=url, **kwargs)

    async def delete(self, url: URLInput, **kwargs) -> JSON:
        """Sends a DELETE request."""
        kwargs.pop("method", None)
        return await self.request("delete", url, **kwargs)

    async def options(self, url: URLInput, **kwargs) -> JSON:
        """Sends an OPTIONS request."""
        kwargs.pop("method", None)
        return await self.request("options", url=url, **kwargs)

    async def head(self, url: URLInput, **kwargs) -> JSON:
        """Sends a HEAD request."""
        kwargs.pop("method", None)
        kwargs.setdefault("allow_redirects", False)
        return await self.request("head", url=url, **kwargs)

    async def patch(self, url: URLInput, **kwargs) -> JSON:
        """Sends a PATCH request."""
        kwargs.pop("method", None)
        return await self.request("patch", url=url, **kwargs)

    def __copy__(self):
        """Do not copy handler"""
        return self

    def __deepcopy__(self, _: dict = None):
        """Do not copy handler"""
        return self



import base64
import hashlib
import logging
import secrets
import sys
import uuid
from abc import ABCMeta
from typing import Any, Self
from urllib.parse import unquote
from webbrowser import open as webopen

from aiohttp import ClientSession
from yarl import URL

from aiorequestful.auth._base import Authoriser, _DEFAULT_SERVICE_NAME
from aiorequestful.auth._utils import AuthRequest, AuthResponseHandler, AuthResponseTester, SocketHandler
from aiorequestful.exception import AuthoriserError
from aiorequestful.types import ImmutableJSON, JSON, URLInput, UnitIterable, Method, Headers
from aiorequestful._utils import get_iterator


class OAuth2Authoriser(Authoriser, metaclass=ABCMeta):

    def __init__(
            self,
            token_request: AuthRequest,
            service_name: str = _DEFAULT_SERVICE_NAME,
            response_handler: AuthResponseHandler = None,
            response_tester: AuthResponseTester = None,
    ):
        super().__init__(service_name=service_name)

        #: Request to exchange the authorisation code for an access token
        self.token_request = token_request

        #: Handles saving and loading token request responses and generates headers from a token request
        self.response_handler = response_handler if response_handler is not None else AuthResponseHandler()
        #: Tests the response given from the token request to ensure the token is valid
        self.response_tester = response_tester if response_tester is not None else AuthResponseTester()

    @staticmethod
    def _encode_client_credentials_as_headers(client_id: str, client_secret: str) -> Headers:
        credentials_encoded = base64.b64encode(f"{client_id}:{client_secret}".encode()).decode()
        return {
            "Authorization": f"Basic {credentials_encoded}",
        }

    def get_response(self) -> JSON | None:
        """Return the loaded the authorisation response if found, or load it from file if available."""
        response = self.response_handler.response
        if not response:
            response = self.response_handler.load_response_from_file()
        return response

    async def _request_token(self, session: ClientSession, request: AuthRequest, params: dict[str, Any] = None) -> JSON:
        with request.enrich_parameters("params", params if params else {}):
            async with request(session=session) as r:
                response = await r.json()

        self.response_handler.enrich_response(response, refresh_token=params.get("refresh_token"))

        sanitised_response = self.response_handler.sanitise_response(response)
        kind = "generated" if not params.get("grant_type") == "refresh_token" else "refreshed"
        self.logger.debug(f"Auth response {kind}: {sanitised_response}")
        return response


class ClientCredentialsFlow(OAuth2Authoriser):

    @classmethod
    def create(
            cls,
            token_request_url: URLInput,
            client_id: str,
            client_secret: str,
            service_name: str = _DEFAULT_SERVICE_NAME,
    ) -> Self:
        """
        Initialises a basic object, generating core request objects from given arguments.
        Sets requests to send the credentials as parameters for each required request.

        :param service_name: The service name for which to authorise.
        :param token_request_url: The URL to call when requesting a new token.
        :param client_id: The client ID.
        :param client_secret: The client secret.
        :return: The initialised object.
        """
        token_request = AuthRequest(
            method=Method.POST,
            url=URL(token_request_url),
            params={
                "client_id": client_id,
                "client_secret": client_secret,
            }
        )

        return ClientCredentialsFlow(
            service_name=service_name,
            token_request=token_request,
        )

    @classmethod
    def create_with_encoded_credentials(
            cls,
            token_request_url: URLInput,
            client_id: str,
            client_secret: str,
            service_name: str = _DEFAULT_SERVICE_NAME,
    ) -> Self:
        """
        Initialises a basic object, generating core request objects from given arguments.
        Encodes the client credentials in base64 format and sets requests to send the encoded credentials
        in the headers of each required request.

        :param service_name: The service name for which to authorise.
        :param token_request_url: The URL to call when requesting a new token.
        :param client_id: The client ID.
        :param client_secret: The client secret.
        :return: The initialised object.
        """
        obj = cls.create(
            token_request_url=token_request_url,
            client_id="",
            client_secret="",
            service_name=service_name,
        )

        credentials_headers = cls._encode_client_credentials_as_headers(
            client_id=client_id, client_secret=client_secret
        )

        if hasattr(obj.token_request, "params"):
            obj.token_request.params.pop("client_id")
            obj.token_request.params.pop("client_secret")
        obj.token_request.headers = credentials_headers

        return obj

    async def authorise(self):
        response = self.get_response()
        loaded = response is not None and response.items()

        valid = await self.response_tester(response=response)

        if not valid:
            if loaded:
                log = "Saved access token not found"
            else:
                log = "Loaded access token is not valid and and no refresh data found"
            self.logger.debug(f"{log}. Generating new token...")

            params = self._generate_request_token_params()
            async with ClientSession() as session:
                response = await self._request_token(session=session, request=self.token_request, params=params)

            valid = await self.response_tester(response=response)

        if not response:
            raise AuthoriserError("Could not generate or load a token")
        if not valid:
            sanitised_response = self.response_handler.sanitise_response(response)
            raise AuthoriserError(f"Auth response is still not valid: {sanitised_response}")

        self.logger.debug("Access token is valid. Saving...")
        self.response_handler.save_response_to_file(response=response)

        return self.response_handler.headers

    @staticmethod
    def _generate_request_token_params() -> dict[str, Any]:
        return {"grant_type": "client_credentials"}


class AuthorisationCodeFlow(OAuth2Authoriser):
    """
    Authorises using OAuth2 specification following the 'Authorization Code' flow specification.

    See more: https://auth0.com/docs/get-started/authentication-and-authorization-flow/authorization-code-flow

    :param service_name: The service name for which to authorise.
    :param user_request: Request to initiate user authentication and authorisation through an `/authorize` endpoint.
    :param token_request: Request to exchange the authorisation code for an access token.
    :param refresh_request: Request to refresh an access token using the refresh token from the token request response.
    :param redirect_uri: The callback URL to apply to the user request to allow
        for the retrieval of the authorisation code.
        Also sent as a parameter to the token request as part of identity confirmation.
    :param socket_handler: Opens a socket on localhost to listen for a request on the redirect_url.
    :param response_handler: Handles manipulation and storing of the response from a token exchange.
    :param response_tester: Tests the response given from the token request to ensure the token is valid.
    """

    @classmethod
    def create(
            cls,
            user_request_url: URLInput,
            token_request_url: URLInput,
            client_id: str,
            client_secret: str,
            refresh_request_url: URLInput = None,
            scope: UnitIterable[str] = (),
            service_name: str = _DEFAULT_SERVICE_NAME,
    ) -> Self:
        """
        Initialises a basic object, generating core request objects from given arguments.
        Sets requests to send the credentials as parameters for each required request.

        :param service_name: The service name for which to authorise.
        :param user_request_url: The URL to configure for the user to authorise access to the application.
        :param token_request_url: The URL to call when requesting a new token.
        :param refresh_request_url: The URL to call when refreshing a token.
        :param client_id: The client ID.
        :param client_secret: The client secret.
        :param scope: The scope/s to request access for from the user during user authorisation.
        :return: The initialised object.
        """
        user_request = AuthRequest(
            method=Method.POST,
            url=URL(user_request_url),
            params={
                "client_id": client_id,
                "scope": " ".join(get_iterator(scope))
            }
        )
        token_request = AuthRequest(
            method=Method.POST,
            url=URL(token_request_url),
            params={
                "client_id": client_id,
                "client_secret": client_secret,
            }
        )
        refresh_request = None if not refresh_request_url else AuthRequest(
            method=Method.POST,
            url=URL(refresh_request_url),
            params={
                "client_id": client_id,
            }
        )

        return AuthorisationCodeFlow(
            service_name=service_name,
            user_request=user_request,
            token_request=token_request,
            refresh_request=refresh_request
        )

    @classmethod
    def create_with_encoded_credentials(
            cls,
            user_request_url: URLInput,
            token_request_url: URLInput,
            client_id: str,
            client_secret: str,
            refresh_request_url: URLInput = None,
            scope: UnitIterable[str] = (),
            service_name: str = _DEFAULT_SERVICE_NAME,
    ) -> Self:
        """
        Initialises a basic object, generating core request objects from given arguments.
        Encodes the client credentials in base64 format and sets requests to send the encoded credentials
        in the headers of each required request.

        :param service_name: The service name for which to authorise.
        :param user_request_url: The URL to configure for the user to authorise access to the application.
        :param token_request_url: The URL to call when requesting a new token.
        :param refresh_request_url: The URL to call when refreshing a token.
        :param client_id: The client ID.
        :param client_secret: The client secret.
        :param scope: The scope/s to request access for from the user during user authorisation.
        :return: The initialised object.
        """
        obj = cls.create(
            service_name=service_name,
            user_request_url=user_request_url,
            token_request_url=token_request_url,
            refresh_request_url=refresh_request_url,
            client_id="",
            client_secret="",
            scope=scope,
        )

        credentials_headers = cls._encode_client_credentials_as_headers(
            client_id=client_id, client_secret=client_secret
        )

        if hasattr(obj.token_request, "params"):
            obj.token_request.params.pop("client_id")
            obj.token_request.params.pop("client_secret")
        obj.token_request.headers = credentials_headers

        if obj.refresh_request:
            if hasattr(obj.refresh_request, "params"):
                obj.refresh_request.params.pop("client_id")
            obj.refresh_request.headers = credentials_headers

        return obj

    def __init__(
            self,
            user_request: AuthRequest,
            token_request: AuthRequest,
            refresh_request: AuthRequest | None = None,
            service_name: str = _DEFAULT_SERVICE_NAME,
            socket_handler: SocketHandler = None,
            redirect_uri: URLInput = None,
            response_handler: AuthResponseHandler = None,
            response_tester: AuthResponseTester = None,
    ):
        super().__init__(
            service_name=service_name,
            token_request=token_request,
            response_handler=response_handler,
            response_tester=response_tester
        )

        #: Request to initiate user authentication and authorisation through an `/authorize` endpoint
        self.user_request = user_request
        #: Request to refresh an access token using the refresh token from the token request response
        self.refresh_request = refresh_request

        if redirect_uri is None:
            redirect_uri = URL.build(scheme="http", host="localhost", port=8080)
        #: The callback URL to apply to the user request to allow for the retrieval of the authorisation code
        self.redirect_uri = redirect_uri

        if socket_handler is None:
            socket_handler = SocketHandler(port=self.redirect_uri.port)
        #: Opens a socket on localhost to listen for a request on the redirect_url
        self.socket_handler = socket_handler

    async def authorise(self):
        response = self.get_response()
        loaded = response is not None and response.items()

        if not loaded:
            self.logger.debug("Saved access token not found. Generating new token...")
            async with ClientSession() as session:
                code = await self._authorise_user(session=session)
                params = self._generate_request_token_params(code=code)
                response = await self._request_token(session=session, request=self.token_request, params=params)

        valid = await self.response_tester(response=response)

        if not valid and loaded:
            response, valid = await self._handle_invalid_loaded_response(response=response)

        if not response:
            raise AuthoriserError("Could not generate or load a token")
        if not valid:
            sanitised_response = self.response_handler.sanitise_response(response)
            raise AuthoriserError(f"Auth response is still not valid: {sanitised_response}")

        self.logger.debug("Access token is valid. Saving...")
        self.response_handler.save_response_to_file(response=response)

        return self.response_handler.headers

    async def _handle_invalid_loaded_response(self, response: ImmutableJSON) -> tuple[JSON, bool]:
        valid = False
        refreshed = False

        async with ClientSession() as session:
            if self.refresh_request and "refresh_token" in response:
                self.logger.debug(
                    "Loaded access token is not valid and refresh data found. Refreshing token and testing..."
                )

                params = self._generate_refresh_token_params(refresh_token=response["refresh_token"])
                response = await self._request_token(session=session, request=self.refresh_request, params=params)
                valid = await self.response_tester(response=response)
                refreshed = True

            if not valid:
                if refreshed:
                    log = "Refreshed access token is still not valid"
                else:
                    log = "Loaded access token is not valid and and no refresh data found"
                self.logger.debug(f"{log}. Generating new token...")

                code = await self._authorise_user(session=session)
                params = self._generate_request_token_params(code=code)
                response = await self._request_token(session=session, request=self.token_request, params=params)

                valid = await self.response_tester(response=response)

        return response, valid

    def _display_message(self, message: str, level: int = logging.INFO) -> None:
        """Log a message and ensure it is displayed to the user no matter the logger configuration."""
        self.logger.log(level=level, msg=message)

        # return if message was logged to stdout
        for handler in self.logger.handlers + list(logging.getHandlerNames()):
            if isinstance(handler, str):
                handler = logging.getHandlerByName(handler)
            if isinstance(handler, logging.StreamHandler) and handler.stream == sys.stdout:
                return

        print(message)

    def _generate_authorise_user_params(self, state: uuid.UUID) -> dict[str, Any]:
        return {"response_type": "code", "redirect_uri": str(self.redirect_uri), "state": str(state)}

    async def _authorise_user(self, session: ClientSession) -> str:
        self.logger.debug("Authorising user privilege access...")

        state = uuid.uuid4()
        params = self._generate_authorise_user_params(state=state)

        with self.socket_handler as listener, self.user_request.enrich_parameters("params", params):
            self._display_message(
                f"\33[1mOpening {self.service_name} in your browser. "
                f"Log in to {self.service_name}, authorise, and return here after \33[0m"
            )
            self._display_message(f"\33[1mWaiting for code, timeout in {listener.timeout} seconds... \33[0m")

            # open authorise webpage and wait for the redirect
            async with self.user_request(session=session) as r:
                webopen(str(r.url))
            request, _ = listener.accept()

            request.send("Code received! You may now close this window".encode("utf-8"))

        self._display_message("\33[92;1mCode received!\33[0m")

        callback_url = URL(
            next(line for line in request.recv(8196).decode().split('\n') if line.startswith("GET"))
        )

        try:
            callback_state = uuid.UUID(unquote(callback_url.query["state"]))
        except ValueError:
            raise AuthoriserError("Returned state is not a valid UUID")

        if callback_state != state:
            raise AuthoriserError("Invalid state returned")

        return unquote(callback_url.query["code"])

    def _generate_request_token_params(self, code: str) -> dict[str, Any]:
        return {"grant_type": "authorization_code", "code": code, "redirect_uri": str(self.redirect_uri)}

    @staticmethod
    def _generate_refresh_token_params(refresh_token: str) -> dict[str, Any]:
        return {"grant_type": "refresh_token", "refresh_token": refresh_token}


class AuthorisationCodePKCEFlow(AuthorisationCodeFlow):
    """
    Authorises using OAuth2 specification following the 'Authorization Code with PKCE' flow specification.

    See more: https://auth0.com/docs/get-started/authentication-and-authorization-flow/authorization-code-flow-with-pkce

    :param service_name: The service name for which to authorise.
    :param user_request: Request to initiate user authentication and authorisation through an `/authorize` endpoint.
    :param token_request: Request to exchange the authorisation code for an access token.
    :param refresh_request: Request to refresh an access token using the refresh token from the token request response.
    :param redirect_uri: The callback URL to apply to the user request to allow
        for the retrieval of the authorisation code.
        Also sent as a parameter to the token request as part of identity confirmation.
    :param socket_handler: Opens a socket on localhost to listen for a request on the redirect_url.
    :param response_handler: Handles manipulation and storing of the response from a token exchange.
    :param response_tester: Tests the response given from the token request to ensure the token is valid.
    """

    @classmethod
    def create(
            cls,
            user_request_url: URLInput,
            token_request_url: URLInput,
            client_id: str,
            refresh_request_url: URLInput = None,
            scope: UnitIterable[str] = (),
            service_name: str = _DEFAULT_SERVICE_NAME,
            **__
    ) -> Self:
        """
        Initialises a basic object, generating core request objects from given arguments.

        :param service_name: The service name for which to authorise.
        :param user_request_url: The URL to configure for the user to authorise access to the application.
        :param token_request_url: The URL to call when requesting a new token.
        :param refresh_request_url: The URL to call when refreshing a token.
        :param client_id: The client ID.
        :param scope: The scope/s to request access for from the user during user authorisation.
        :return: The initialised object.
        """
        user_request = AuthRequest(
            method=Method.POST,
            url=URL(user_request_url),
            params={
                "client_id": client_id,
                "scope": " ".join(get_iterator(scope)),
            }
        )
        token_request = AuthRequest(
            method=Method.POST,
            url=URL(token_request_url),
            params={
                "client_id": client_id,
            }
        )
        refresh_request = None if not refresh_request_url else AuthRequest(
            method=Method.POST,
            url=URL(refresh_request_url),
            params={
                "client_id": client_id,
            }
        )

        return AuthorisationCodePKCEFlow(
            service_name=service_name,
            user_request=user_request,
            token_request=token_request,
            refresh_request=refresh_request
        )

    def __init__(
            self,
            user_request: AuthRequest,
            token_request: AuthRequest,
            refresh_request: AuthRequest | None = None,
            service_name: str = _DEFAULT_SERVICE_NAME,
            socket_handler: SocketHandler = None,
            redirect_uri: URLInput = None,
            response_handler: AuthResponseHandler = None,
            response_tester: AuthResponseTester = None,
            pkce_code_length: int = 128,
    ):
        if not 43 <= pkce_code_length <= 128:
            raise AuthoriserError("PKCE code length must be between 43 and 128 inclusive")
        self.code_verifier = secrets.token_urlsafe(96)[:pkce_code_length]

        super().__init__(
            service_name=service_name,
            user_request=user_request,
            token_request=token_request,
            refresh_request=refresh_request,
            socket_handler=socket_handler,
            redirect_uri=redirect_uri,
            response_handler=response_handler,
            response_tester=response_tester,
        )

    def _generate_authorise_user_params(self, state: uuid.UUID) -> dict[str, Any]:
        code_hashed = hashlib.sha256(self.code_verifier.encode("ascii")).digest()
        code_encoded = base64.urlsafe_b64encode(code_hashed)
        code_challenge = code_encoded.decode("ascii")[:-1]

        return {
            "response_type": "code",
            "redirect_uri": str(self.redirect_uri),
            "state": str(state),
            "code_challenge_method": "S256",
            "code_challenge": code_challenge,
        }

    def _generate_request_token_params(self, code: str) -> dict[str, Any]:
        return {
            "grant_type": "authorization_code",
            "code": code,
            "redirect_uri": str(self.redirect_uri),
            "code_verifier": self.code_verifier
        }



from collections.abc import Iterator, Iterable
from typing import Any

from yarl import URL

from aiorequestful.exception import AIORequestsImportError
from aiorequestful.types import UnitIterable


def get_iterator(value: Any) -> Iterator | None:
    """
    Safely get an iterator.

    When a string is given, the iterator will be size 1 where the first element is the given string.
    """
    if value is None:
        return iter(())
    elif isinstance(value, Iterator):
        return value
    elif isinstance(value, str) or not isinstance(value, Iterable):
        value = (value,)
    return iter(value)


def format_url_log(method: str, url: URL, messages: UnitIterable[Any]) -> str:
    """Format a request for a given ``url`` of a given ``method`` appending the given ``messages``"""
    url = str(url.with_query(None))
    url_pad_map = [30, 40, 70, 100]
    url_pad = next((pad for pad in url_pad_map if len(url) < pad), url_pad_map[-1])

    return f"{method.upper():<7}: {url:<{url_pad}} | {" | ".join(map(str, get_iterator(messages)))}"


def required_modules_installed(modules: list, this: object = None) -> bool:
    """Check the required modules are installed, raise :py:class:`MusifyImportError` if not."""
    modules_installed = all(module is not None for module in modules)
    if not modules_installed and this is not None:
        names = [name for name, obj in globals().items() if obj in modules and not name.startswith("_")]
        if isinstance(this, str):
            message = f"Cannot run {this}. Required modules: {", ".join(names)}"
        else:
            message = f"Cannot create {this.__class__.__name__} object. Required modules: {", ".join(names)}"

        raise AIORequestsImportError(message)

    return modules_installed
